'''
RLP Encoding/Decoding.

RLP encodes before storing on disk or transmitting on network.

Primary RLP can only deal with "item" type,
The definition of item is:
1) byte string (bytes in Python)
2) list

Some exmples are:
- b'\x00\xff'
- empty list []
- list of bytes [ b'\x00', b'\x01\x03']
- list of combinations [ [], b'\x00', [b'\x00']]

The encoded result is bytes. The encoded methods is called RLP.

         RLP    +-----------+
 item +-------> |RPL encoded|
                +-----------+

But in the real world, the inputs are not pure bytes nor lists.
Some are of complex key-value pairs like dict.
Some are of "0x123" form of number.

This module exists for some pre-defined
real world object => "item" conversion.

                         serialize
 "real world object" +--------------> item

'''
from typing import Tuple
from typing import Union
from typing import List
from typing import Any
import re
from rlp.sedes import BigEndianInt
from rlp.exceptions import DeserializationError, SerializationError
from rlp import encode as rlp_encode
from rlp import decode as rlp_decode


def _is_hex_string(a: str, must_contain_data: bool) -> bool:
    c = None
    if must_contain_data:
        c = re.compile('^0x[0-9a-f]+$', re.I)
    else:
        c = re.compile('^0x[0-9a-f]*$', re.I)

    if c.match(a):
        return True
    else:
        return False


def _is_decimal_string(a: str) -> bool:
    c = re.compile('^[0-9]+$')
    if c.match(a):
        return True
    else:
        return False


def _is_pure_int(a: int) -> bool:
    return type(a) == int


def _is_pure_str(a: str) -> bool:
    return type(a) == str


class ScalarKind():
    pass


class BytesKind(ScalarKind):
    '''
    Convert bytes type of Python object to RLP "item".
    '''
    @classmethod
    def is_valid_type(cls, obj):
        return isinstance(obj, (bytes, bytearray))

    def serialize(self, obj: bytes) -> bytes:
        '''
        Serialize the object into a RLP encode-able "item".

        Parameters
        ----------
        obj : bytes
            The input.

        Returns
        -------
        bytes
            The "item" in bytes.

        Raises
        ------
        SerializationError
            raise if input is not bytes.
        '''
        if not self.is_valid_type(obj):
            raise SerializationError(
                'type of "obj" param is not right, bytes required.', obj)

        return obj

    def deserialize(self, serial: bytes) -> bytes:
        '''
        De-serialize a RLP "item" back to bytes.

        Parameters
        ----------
        serial : bytes
            The input.

        Returns
        -------
        bytes
            Original bytes.

        Raises
        ------
        DeserializationError
            raise if input is not bytes.
        '''
        if not self.is_valid_type(serial):
            raise DeserializationError(
                'type of "serial" param is not right, bytes required.', serial)

        return serial


class NumericKind(ScalarKind, BigEndianInt):
    '''
    This is a pre-defined type for Number-like objects.

    Good examples are:
    '0x0', '0x123', '0', '100', 0, 0x123

    Bad examples are:
    '0x123z', {}, '0x', -1, '0x12345678123456780', 2 ** 64
    '''

    def __init__(self, max_bytes: int = None):
        '''
        Initialize a NumericKind.

        Parameters
        ----------
        max_bytes : int
            Max bytes in the encoded result. (not enough then prepend 0)
        '''
        self.max_bytes = max_bytes
        super().__init__(l=max_bytes)

    def serialize(self, obj: Union[str, int]) -> bytes:
        '''
        Serialize the object into a RLP encode-able "item".

        Parameters
        ----------
        obj : Union[str, int]
            obj is either number in string or number int.
        '''

        if not (_is_pure_str(obj) or _is_pure_int(obj)):
            raise SerializationError("The input is not str nor int.", obj)

        number = None

        if _is_pure_str(obj):
            if _is_hex_string(obj, True):
                number = int(obj, 16)

            if _is_decimal_string(obj):
                number = int(obj)

        if _is_pure_int(obj):
            number = obj

        # remove leading 0 from bytes sequence.
        result_bytes = super().serialize(number)
        byte_list = []
        can_append_flag = False
        for x in result_bytes:
            if not can_append_flag:
                if x != 0:
                    can_append_flag = True
                else:
                    continue

            if can_append_flag:
                byte_list.append(x)

        return bytes(byte_list)

    def deserialize(self, serial) -> int:
        '''
        Deserialize bytes to int.

        Parameters
        ----------
        serial : [type]
            bytes

        Returns
        -------
        int
            integer

        Raises
        ------
        DeserializationError
            If bytes contain leading 0.
        '''
        if len(serial) > 0 and serial[0] == 0:
            raise DeserializationError(
                "Leading 0 should be removed from bytes",
                serial
            )

        # add leading 0 to bytes sequence if width is set.
        if self.max_bytes:
            byte_list = [x for x in serial]
            length = len(byte_list)
            missed = self.max_bytes - length
            if missed:
                byte_list = [0] * missed + byte_list
            serial2 = bytes(byte_list)
        else:
            serial2 = serial
        return super().deserialize(serial2)


class BlobKind(ScalarKind):
    '''
    This is a pre-defined type for '0x....' like hex strings,
    which shouldn't be interpreted as a number, usually an identifier.

    like: address, block_ref, data to smart contract.
    '''

    def serialize(self, obj: str) -> bytes:
        '''
        Serialize a '0x...' string to bytes.

        Parameters
        ----------
        obj : str
            '0x...' style string.

        Returns
        -------
        bytes
            the "item" that can be rlp encodeded.
        '''
        if not _is_hex_string(obj, False):
            raise SerializationError('expect 0x... style string', obj)

        if len(obj) % 2 != 0:
            raise SerializationError(
                'expect 0x... style string of even length.', obj)

        obj2 = obj[2:]  # remove '0x'

        return bytes.fromhex(obj2)

    def deserialize(self, serial: bytes) -> str:
        '''
        Deserialize bytes to '0x...' string.

        Parameters
        ----------
        serial : bytes
            the bytes.

        Returns
        -------
        str
            string of style '0x...'
        '''

        return '0x' + serial.hex()


class FixedBlobKind(BlobKind):
    '''
    This is a pre-defined type for '0x....' like hex strings,
    which shouldn't be interpreted as a number, usually an identifier.

    like: address, block_ref, data to smart contract.

    Note
    ----
        This kind has a fixed length of bytes.
        (also means the input hex is fixed length)
    '''

    def __init__(self, byte_length):
        self.byte_length = byte_length

    def serialize(self, obj: str) -> bytes:
        # 0x counts for 2 chars. 1 bytes = 2 hex char.
        allowed_hex_length = self.byte_length * 2 + 2

        if len(obj) != allowed_hex_length:
            raise SerializationError(
                "Max allowed string length {}".format(allowed_hex_length),
                obj
            )

        return super().serialize(obj)

    def deserialize(self, serial: bytes) -> str:
        if len(serial) != self.byte_length:
            raise DeserializationError(
                "Bytes should be {} long.".format(self.byte_length),
                serial
            )

        return super().deserialize(serial)


class NoneableFixedBlobKind(FixedBlobKind):
    '''
    This is a pre-defined type for '0x....' like hex strings,
    which shouldn't be interpreted as a number, usually an identifier.

    like: address, block_ref, data to smart contract.

    Note
    ----
        This kind has a fixed length of bytes.
        (also means the input hex is fixed length)

        For this kind, input can be None.
        Then decoded is also None.
    '''

    def __init__(self, byte_length):
        super().__init__(byte_length)

    def serialize(self, obj: str = None) -> bytes:
        if obj is None:
            return bytes(0)

        return super().serialize(obj)

    def deserialize(self, serial: bytes) -> str:
        if len(serial) == 0:
            return None

        return super().deserialize(serial)


class CompactFixedBlobKind(FixedBlobKind):
    '''
    This is a pre-defined type for '0x....' like strings,
    which shouldn't be interpreted as a number, usually an identifier.

    like: address, block_ref, data to smart contract.

    Note
    ----
        When encode, the result fixed length bytes will be
        removed of leading zeros. i.e. 000123 -> 123

        When decode, it expects the input bytes length <= fixed_length.
        and it pads the leading zeros back. Output '0x{0}paddingxxx...'
    '''

    def __init__(self, byte_length):
        super().__init__(byte_length)

    def serialize(self, obj: str) -> bytes:
        b = super().serialize(obj)
        b_list = [x for x in b if x != 0]
        if (len(b_list) == 0):
            return bytes(0)
        else:
            return bytes(b_list)

    def deserialize(self, serial: bytes) -> str:
        if (len(serial) > self.byte_length):
            raise DeserializationError(
                "Bytes too long, only need {}".format(self.byte_length),
                serial
            )

        if len(serial) == 0 or serial[0] == 0:
            raise DeserializationError(
                "No leading zeros. And byte sequence length should be > 0",
                serial
            )

        missing = self.byte_length - len(serial)
        b_list = [0] * missing + [x for x in serial]
        return super().deserialize(bytes(b_list))


class BaseWrapper():
    ''' BaseWrapper is a container for complex types to be encode/decoded. '''
    pass


class DictWrapper(BaseWrapper):
    ''' DictWrapper is a container for parsing dict like objects. '''

    def __init__(self, list_of_tuples: List[Tuple[str, Union[BaseWrapper, ScalarKind]]]):
        '''Constructor

        Parameters
        ----------
        list_of_tuples : List[Tuple[str, Union[BaseWrapper, ScalarKind]]]
            A list of tuples.
            eg. [(key, codec), (key, codec) ... ])
            key is a string.
            codec is either a BaseWrapper, or a ScalarKind.
        '''
        self.keys = [x[0] for x in list_of_tuples]
        self.codecs = [x[1] for x in list_of_tuples]


class ListWrapper(BaseWrapper):
    '''
    ListWrapper is a container for parsing a list,
    the items type in the list can be heterogeneous.
    '''

    def __init__(self, list_of_codecs: List[Union[BaseWrapper, ScalarKind]]):
        '''Constructor

        Parameters
        ----------
        list_of_codecs : List[Union[BaseWrapper, ScalarKind]]
            A list of codecs.
            eg. [codec, codec, codec...]
            codec is either a BaseWrapper, or a ScalarKind.
        '''
        self.codecs = list_of_codecs


class HomoListWrapper(BaseWrapper):
    '''
    HomoListWrapper is a container for parsing a list,
    the items in the list are of the same type.
    '''

    def __init__(self, codec: Union[BaseWrapper, ScalarKind]):
        '''Constructor

        Parameters
        ----------
        list_of_codecs : List[Union[BaseWrapper, ScalarKind]]
            A list of codecs.
            eg. [codec, codec, codec...]
            codec is either a BaseWrapper, or a ScalarKind.
        '''
        self.codec = codec


def pack(obj, wrapper: Union[BaseWrapper, ScalarKind]) -> Union[bytes, List]:
    '''Pack a Python object according to wrapper.

    Parameters
    ----------
    obj : Any
        A dict, a list, or a string/int/any...
    wrapper : Union[BaseWrapper, ScalarKind]
        A Wrapper.

    Returns
    -------
    Union[bytes, List]
        Returns either the bytes if obj is a basic type,
        or a list if obj is dict/list.

    Raises
    ------
    Exception
        If the wrapper/codec is unknown.
    '''
    # Simple wrapper: ScalarKind
    if isinstance(wrapper, ScalarKind):
        return wrapper.serialize(obj)

    # Complicated wrapper: BaseWrapper
    if isinstance(wrapper, BaseWrapper):
        if isinstance(wrapper, DictWrapper):
            r = []
            for (key, codec) in zip(wrapper.keys, wrapper.codecs):
                r.append(pack(obj[key], codec))
            return r

        if isinstance(wrapper, ListWrapper):
            r = []
            for (item, codec) in zip(obj, wrapper.codecs):
                r.append(pack(item, codec))
            return r

        if isinstance(wrapper, HomoListWrapper):
            r = []
            for item in obj:
                r.append(pack(item, wrapper.codec))
            return r

        raise Exception('codec type is unknown.')

    # Wrapper type is unknown, raise.
    raise Exception('wrapper type is unknown.{}'.format(wrapper))


def unpack(packed: Union[List, bytes], wrapper: Union[BaseWrapper, ScalarKind]) -> Union[dict, List, Any]:
    '''Unpack a serialized thing back into a dict/list or a Python basic type.

    Parameters
    ----------
    packed : Union[List, bytes]
        A list of RLP encoded or pure bytes.
    wrapper : Union[BaseWrapper, ScalarKind]
        The Wrapper.

    Returns
    -------
    Union[dict, List, Any]
        dict/list if the wrapper instruction is dict/list,
        Python basic type if input is bytes.

    Raises
    ------
    Exception
        If the wrapper/codec is unknown.
    '''
    # Simple wrapper: ScalarKind
    if isinstance(wrapper, ScalarKind):
        return wrapper.deserialize(packed)

    # Complicated wrapper: BaseWrapper
    if isinstance(wrapper, BaseWrapper):
        if isinstance(wrapper, DictWrapper):
            r = {}
            for (blob, key, codec) in zip(packed, wrapper.keys, wrapper.codecs):
                r[key] = unpack(blob, codec)
            return r

        if isinstance(wrapper, ListWrapper):
            r = []
            for (blob, codec) in zip(packed, wrapper.codecs):
                r.append(unpack(blob, codec))
            return r

        if isinstance(wrapper, HomoListWrapper):
            r = []
            for blob in packed:
                r.append(unpack(blob, wrapper.codec))
            return r

        raise Exception('codec type is unknown.')

    # Wrapper type is unknown, raise.
    raise Exception('wrapper type is unknown.')


class ComplexCodec(object):
    def __init__(self, wrapper: BaseWrapper):
        self.wrapper = wrapper

    def encode(self, data: Any) -> bytes:
        packed = pack(data, self.wrapper)
        return rlp_encode(packed)

    def decode(self, data: bytes):
        to_be_unpacked = rlp_decode(data)
        return unpack(to_be_unpacked, self.wrapper)



''' 
User signed certificate.

https://github.com/vechain/VIPs/blob/master/vips/VIP-192.md
'''
from typing import Optional
import json
import re
import copy
from .cry import blake2b256
from .cry import secp256k1
from .cry import address


class Certificate():
    def __init__(
        self,
        purpose: str,
        payload: dict,
        domain: str,
        timestamp: int,
        signer: str,
        signature: Optional[str] = None
    ):
        '''
        Certficate itself.

        Parameters
        ----------
        purpose : str
            A String.
        payload : dict
            Of style { "type": str, "content": str}
        domain : str
            A String
        timestamp : int
            Integer, Unix timestamp.
        signer : str
            0x... the signer address.
        signature : Optional[str], optional
            A secp256k1 signed bytes, but turned into a '0x' + bytes.hex() format, by default None
        '''
        if not payload.get('type'):
            raise ValueError('payload needs a string field "type"')
        if not payload.get('content'):
            raise ValueError('payload needs a string field "content"')

        self.obj = {
            'purpose': purpose,
            'payload': payload,
            'domain': domain,
            'timestamp': timestamp,
            'signer': signer
        }

        if signature:
            self.obj['signature'] = signature

    def to_dict(self):
        return self.obj


def safe_tolowercase(s: str):
    if type(s) == str:
        return s.lower()
    else:
        return s


def encode(cert: Certificate) -> str:
    '''
    Encode a certificate into json.

    Parameters
    ----------
    cert : Certificate
        The certificate to be encoded.

    Returns
    -------
    str
        The encoded string.
    '''
    temp = cert.to_dict()
    temp['signer'] = safe_tolowercase(temp['signer'])
    if temp.get('signature'):
        temp['signature'] = safe_tolowercase(temp['signature'])

    # separators=(',', ':') -> no whitespace compact string
    # sort_keys -> dict key is ordered.
    return json.dumps(temp, separators=(',', ':'), sort_keys=True)


SIGNATURE_PATTERN = re.compile('^0x[0-9a-f]+$', re.I)


def verify(cert: Certificate):
    temp = cert.to_dict()
    if not temp.get('signature'):
        raise ValueError('Cert needs a "signature" field.')

    sig = copy.copy(temp['signature'])
    if len(sig) % 2 != 0:
        raise ValueError('Cert "signature" field needs to be of even length.')

    if not SIGNATURE_PATTERN.match(sig):
        raise ValueError('Cert "signature" field can not pass the style check')

    # remove the signature, then encode.
    del temp['signature']
    the_encoded = encode(Certificate(**temp))
    signing_hash, _ = blake2b256([the_encoded.encode('utf-8')])
    pub_key = secp256k1.recover(signing_hash, bytes.fromhex(sig[2:]))

    if '0x' + address.public_key_to_address(pub_key).hex() != safe_tolowercase(temp['signer']):
        raise Exception('signature does not match with the signer.')



'''
Keccak

Keccak hash function.
'''

import sha3  # pysha3
from typing import List, Tuple


def keccak256(list_of_bytes: List[bytes]) -> Tuple[bytes, int]:
    '''
    Compute the sha3_256 flavor hash, outputs 256 bits / 32 bytes.

    Parameters
    ----------
    list_of_bytes : List[bytes]
        A list of bytes to be hashed.

    Returns
    -------
    Tuple[bytes, int]
        Hash value in bytes and length of bytes.
    '''
    m = sha3.keccak_256()
    for item in list_of_bytes:
        m.update(item)

    return m.digest(), m.digest_size



'''
Bloom Filter.

A data structure tells us either the element definitely is not in,
or may be in the set.

Instead of a traditional hash-based set takes up too much memory,
this structure permits less memory with a tolerable false positive rate.

m = total bits of the filter.
k = how many different hash functions to use.
n = number of elements to be added to the filter.

2048 bits / 256 bytes
'''
import math
from typing import Callable
from .cry import blake2b256


class Bloom:
    MAX_K = 16
    BITS_LENGTH = 2048

    @classmethod
    def estimate_k(cls, count: int) -> int:
        '''
        Estimate the k based on the number of elements
        to be inserted into bloom filter.

        Parameters
        ----------
        count : int
            The number of elements to be inserted.

        Returns
        -------
        int
            The estimated k.
        '''
        k = round(cls.BITS_LENGTH / count * math.log(2))
        return max(min(k, cls.MAX_K), 1)

    def __init__(self, k: int, bits: bytes = None):
        '''
        Construct a bloom filter.
        k is the number of different hash functions.


        Parameters
        ----------
        k : int
            The number of different hash functions to use.
        bits : bytes, optional
            previous bloom filter to inherit, by default None.
            Leave it None to create an empty bloom filter.
        '''
        self.k = k
        if bits is None:
            self.bits = bytes(self.BITS_LENGTH//8)
        else:
            self.bits = bits

    def _distribute(self, element: bytes, tester: Callable[[int, int], bool]) -> bool:
        '''
        Distribute the element into the bloom filter.

        Parameters
        ----------
        element : bytes
            the element to be fit into the bloom filter.
        tester : Callable[[int, int], bool]
            a function to test the bit, return False to stop the operation.

        Returns
        -------
        bool
            True/False if element is inside during testing,
            or True when adding element.
        '''
        h, h_length = blake2b256([element])
        for x in range(0, self.k):
            d = (h[x * 2 + 1] + (h[x * 2] << 8)) % self.BITS_LENGTH
            bit = 1 << (d % 8)
            if not tester(int(d / 8), bit):
                return False
        return True

    def add(self, element: bytes) -> bool:
        '''
        Add an element to the bloom filter.

        Parameters
        ----------
        element : bytes
            The element in bytes.

        Returns
        -------
        bool
            True
        '''
        def t(index: int, bit: int):
            temp = list(self.bits)
            temp[index] = temp[index] | bit
            self.bits = bytes(temp)
            return True

        return self._distribute(element, t)

    def test(self, element: bytes) -> bool:
        '''
        Test if element is inside the bloom filter.

        Parameters
        ----------
        element : bytes
            The element in bytes.

        Returns
        -------
        bool
            True if inside, False if not inside.
        '''
        def t(index: int, bit: int):
            return (self.bits[index] & bit) == bit

        return self._distribute(element, t)



'''
HD nodes, HD wallets.

Hierarchically Deterministic Wallets for VeChain.

Relevant information: BIP32 and BIP44.
BIP32: https://github.com/bitcoin/bips/blob/master/bip-0032.mediawiki
BIP44: https://github.com/bitcoin/bips/blob/master/bip-0044.mediawiki

BIP-44 specified path notation:
m / purpose' / coin_type' / account' / change / address_index

Derive path for the VET:
m / 44' / 818' / 0' / 0 /<address_index>

So the following is the root of the "external" node chain for VET.

m / 44' / 818' / 0' / 0

m is the master key, which shall be generated from a seed.

The following is the "first" key pair on the "external" node chain.

m / 44' / 818' / 0' / 0 / 0

'''
from typing import List
from .mnemonic import derive_seed
from .address import public_key_to_address
from bip_utils import Bip32, Bip32Utils, Base58Encoder
from eth_keys import KeyAPI


VET_EXTERNAL_PATH = "m/44'/818'/0'/0"

VERSION_MAINNET_PUBLIC = bytes.fromhex('0488B21E')
VERSION_MAINNET_PRIVATE = bytes.fromhex('0488ADE4')
DEPTH_MASTER_NODE = bytes.fromhex('00')
FINGER_PRINT_MASTER_KEY = bytes.fromhex('00000000')
CHILD_NUMBER_MASTER_KEY = bytes.fromhex('00000000')


def _strip_0x04(p: bytes):
    if len(p) == 65:
        return p[1:]
    else:
        return p


class HDNode():
    '''
    HD Node that is able to derive child HD Node.

    Please use static methods provided in this class to construct
    new instances rather than instantiate one by hand.
    '''

    def __init__(self, bip32_ctx: Bip32):
        '''
        HDNode constructor, it is not recommended to use this directly.
        To construct an HDNode, use staticmethods below instead.

        Parameters
        ----------
        bip32_ctx : Bip32
        '''

        self.bip32_ctx = bip32_ctx

    @staticmethod
    def from_seed(seed: bytes, init_path=VET_EXTERNAL_PATH):
        '''
        Construct an HD Node from a seed (64 bytes).
        The init_path is m/44'/818'/0'/0 for starting.
        or you can simply put in 44'/818'/0'/0

        Note
        ----
            The seed will be further developed into
            a "m" secret key and "chain code".

        Parameters
        ----------
        seed : bytes
            Seed itself.
        init_path : str, optional
            The derive path, by default VET_EXTERNAL_PATH

        Returns
        -------
        HDNode
            A new HDNode.
        '''
        bip32_ctx = Bip32.FromSeedAndPath(seed, init_path)
        return HDNode(bip32_ctx)

    @staticmethod
    def from_mnemonic(words: List[str], init_path=VET_EXTERNAL_PATH):
        '''
        Construct an HD Node from a set of words.
        The init_path is m/44'/818'/0'/0 by default on VeChain.

        Note
        ----
            The words will generate a seed,
            which will be further developed into
            a "m" secret key and "chain code".

        Parameters
        ----------
        words : List[str]
            Mnemonic words, usually 12 words.
        init_path : str, optional
            The initial derivation path, by default VET_EXTERNAL_PATH

        Returns
        -------
        HDNode
            A new HDNode.
        '''

        seed = derive_seed(words)  # 64 bytes
        bip32_ctx = Bip32.FromSeedAndPath(seed, init_path)
        return HDNode(bip32_ctx)

    @staticmethod
    def from_public_key(pub: bytes, chain_code: bytes):
        '''
        Construct an HD Node from an uncompressed public key.
        (starts with 0x04 as first byte)

        Parameters
        ----------
        pub : bytes
            An uncompressed public key in bytes.
        chain_code : bytes
            32 bytes

        Returns
        -------
        HDNode
            A new HDNode.
        '''
        # parts
        net_version = VERSION_MAINNET_PUBLIC
        depth = DEPTH_MASTER_NODE
        fprint = FINGER_PRINT_MASTER_KEY
        index = CHILD_NUMBER_MASTER_KEY
        chain = chain_code
        key_bytes = KeyAPI.PublicKey(_strip_0x04(pub)).to_compressed_bytes()

        # assemble
        all_bytes = net_version + depth + fprint + index + chain + key_bytes
        # double sha-256 checksum
        xpub_str = Base58Encoder.CheckEncode(all_bytes)
        bip32_ctx = Bip32.FromExtendedKey(xpub_str)
        return HDNode(bip32_ctx)

    @staticmethod
    def from_private_key(priv: bytes, chain_code: bytes):
        '''
        Construct an HD Node from a private key.

        Parameters
        ----------
        priv : bytes
            The privte key in bytes.
        chain_code : bytes
            32 bytes of random number you choose.

        Returns
        -------
        HDNode
            A new HDNode.
        '''

        # print('input priv', len(priv))
        # parts
        net_version = VERSION_MAINNET_PRIVATE
        depth = DEPTH_MASTER_NODE
        fprint = FINGER_PRINT_MASTER_KEY
        index = CHILD_NUMBER_MASTER_KEY
        chain = chain_code
        key_bytes = b'\x00' + priv

        # assemble
        all_bytes = net_version + depth + fprint + index + chain + key_bytes
        # double sha-256 checksum
        xpriv = Base58Encoder.CheckEncode(all_bytes)
        bip32_ctx = Bip32.FromExtendedKey(xpriv)

        return HDNode(bip32_ctx)

    def derive(self, index: int):
        '''
        Derive the child HD Node from current HD Node.

        Note
        ----
            private key -> private key.
            private key -> public key.
            public key -> public key. 
            public key -> private key. (CAN NOT!)

        Parameters
        ----------
        index : int
            Which key index (0,1,2... 2^32-1) to derive.

        Returns
        -------
        HDNode
            A New HDNode.
        '''

        bip32_ctx = self.bip32_ctx.DerivePath(str(index))

        return HDNode(bip32_ctx)

    def public_key(self) -> bytes:
        '''
        Get current node's public key in uncompressed format bytes.
        (starts with 0x04)

        Returns
        -------
        bytes
            The uncompressed public key.
        '''
        return b'\x04' + self.bip32_ctx.PublicKey().RawUncompressed().ToBytes()

    def private_key(self) -> bytes:
        '''
        Get current node's private key in bytes format.
        If this node was publicly derived,
        then call this function may cause a Bip32KeyError exception.

        Returns
        -------
        bytes
            The private key in bytes.
        '''
        return self.bip32_ctx.PrivateKey().Raw().ToBytes()

    def chain_code(self) -> bytes:
        '''
        Get the chaincode of current HD node.

        Returns
        -------
        bytes
            32 bytes of chain code.
        '''
        return self.bip32_ctx.Chain()

    def address(self) -> bytes:
        '''
        Get the common address format.

        Returns
        -------
        bytes
            The address in bytes. (without prefix 0x)
        '''
        return public_key_to_address(self.public_key())

    def finger_print(self) -> bytes:
        '''
        Get the finger print of current HD Node public key.

        Returns
        -------
        bytes
            finger print in bytes.
        '''
        return self.bip32_ctx.FingerPrint()



'''
secp256k1 Elliptic Curve related functions.

1) Generate a private Key.
2) Derive uncompressed public key from private key.
3) Sign a message hash using the private key, generate signature.
4) Given the message hash and signature, recover the uncompressed public key.
'''

from ecdsa import SigningKey, SECP256k1
from eth_keys import KeyAPI


MAX = bytes.fromhex(
    'fffffffffffffffffffffffffffffffebaaedce6af48a03bbfd25e8cd0364141')
ZERO = bytes.fromhex('0' * 64)


def _is_valid_private_key(priv_key: bytes) -> bool:
    '''
    Verify if a private key is good.

    Returns
    -------
    bool
        True if the private key is valid.
    '''
    if priv_key == ZERO:
        return False

    if priv_key >= MAX:
        return False

    return True


def _is_valid_message_hash(msg_hash: bytes) -> bool:
    '''
    Verify if a message hash is in correct format.
    (as in terms of VeChain)

    Parameters
    ----------
    msg_hash: bytes
        The msg hash to be processed.

    Returns
    -------
    bool
        If the message hash is in correct format or not.
    '''
    return len(msg_hash) == 32


def generate_privateKey() -> bytes:
    '''
    Create a random number(32 bytes) as private key.

    Returns
    -------
    bytes
        The private key in 32 bytes format.
    '''
    while True:
        _a = SigningKey.generate(curve=SECP256k1)
        if _is_valid_private_key(_a.to_string()):
            return _a.to_string()


def derive_publicKey(priv_key: bytes) -> bytes:
    '''
    Derive public key from a private key(uncompressed).

    Parameters
    ----------
    priv_key: bytes
        The private key in bytes.

    Returns
    -------
    bytes
        The public key(uncompressed) in bytes,
        which starts with 04.

    Raises
    ------
    ValueError
        If the private key is not valid.
    '''
    if not _is_valid_private_key(priv_key):
        raise ValueError('Private Key not valid.')

    _a = SigningKey.from_string(priv_key, curve=SECP256k1)
    return _a.verifying_key.to_string("uncompressed")


def sign(msg_hash: bytes, priv_key: bytes) -> bytes:
    '''
    Sign the message hash.
    (not the message itself)

    Parameters
    ----------
    msg_hash: bytes
        The message hash.
    priv_key: bytes
        The private key in bytes.

    Returns
    -------
    bytes
        The signing result.

    Raises
    ------
    ValueError
        If the input is malformed.
    '''

    if not _is_valid_private_key(priv_key):
        raise ValueError('Private Key not valid.')

    sig = KeyAPI().ecdsa_sign(msg_hash, KeyAPI.PrivateKey(priv_key))

    r = sig.r.to_bytes(32, byteorder='big')
    s = sig.s.to_bytes(32, byteorder='big')
    v = sig.v.to_bytes(1, byteorder='big')  # public key recovery bit.

    return b''.join([r, s, v])  # 32 + 32 + 1 bytes


def recover(msg_hash: bytes, sig: bytes) -> bytes:
    '''
    Recover the uncompressed public key from signature.

    Parameters
    ----------
    msg_hash: bytes
        The message hash.
    sig: bytes
        The signature.

    Returns
    -------
    bytes
        public key in uncompressed format.

    Raises
    ------
    ValueError
        If the signature is bad,
        or recovery bit is bad,
        or cannot recover(sig and msg_hash doesn't match).
    '''

    if not _is_valid_message_hash(msg_hash):
        raise ValueError('Message Hash must be 32 bytes.')

    if len(sig) != 65:
        raise ValueError('Signature must be 65 bytes.')

    if not (sig[64] == 0 or sig[64] == 1):
        raise ValueError('Signature last byte must be 0 or 1')

    pk = KeyAPI().ecdsa_recover(
        msg_hash,
        KeyAPI.Signature(signature_bytes=sig)
    )

    # uncompressed should have first byte = 04
    return bytes([4]) + pk.to_bytes()



'''
Transaction class defines VeChain's multi-clause transaction (tx).

This module defines data structure of a tx, and the encoding/decoding of tx data.
'''
from voluptuous import Schema, Any, Optional, REMOVE_EXTRA
from typing import Union, List
from copy import deepcopy
from .rlp import NumericKind, CompactFixedBlobKind, NoneableFixedBlobKind, BlobKind, BytesKind
from .rlp import DictWrapper, HomoListWrapper
from .rlp import ComplexCodec
from .cry import blake2b256
from .cry import secp256k1
from .cry import address

# Kind Definitions
# Used for VeChain's "reserved features" kind.
FeaturesKind = NumericKind(4)

# Unsigned/Signed RLP Wrapper.
_params = [
    ("chainTag", NumericKind(1)),
    ("blockRef", CompactFixedBlobKind(8)),
    ("expiration", NumericKind(4)),
    ("clauses", HomoListWrapper(codec=DictWrapper([
        ("to", NoneableFixedBlobKind(20)),
        ("value", NumericKind(32)),
        ("data", BlobKind())
    ]))),
    ("gasPriceCoef", NumericKind(1)),
    ("gas", NumericKind(8)),
    ("dependsOn", NoneableFixedBlobKind(32)),
    ("nonce", NumericKind(8)),
    ("reserved", HomoListWrapper(codec=BytesKind()))
]

# Unsigned Tx Wrapper
UnsignedTxWrapper = DictWrapper(_params)

# Signed Tx Wrapper
SignedTxWrapper = DictWrapper( _params + [("signature", BytesKind())] )

CLAUSE = Schema(
    {
        "to": Any(str, None), # Destination contract address, or set to None to create contract.
        "value": Any(str, int), # VET to pass to the call.
        "data": str
    },
    required=True,
    extra=REMOVE_EXTRA
)

# class Clause():
#     '''
#     Clause type.
#     Consists of the "destination", the "vet value" to pass to, and the "data" to pass to.
#     '''

#     def __init__(
#         self,
#         to: Union[str, None],
#         value: Union[str, int],
#         data: str
#     ):
#         '''
#         Create a clause.

#         Parameters
#         ----------
#         to : Union[str, None]
#             Destination contract address, or set to None to create contract.
#         value : Union[str, int]
#             VET to pass to the call.
#         data : str
#             data for contract method invocation or deployment.
#         '''
#         self.to = to
#         self.value = value
#         self.data = data

#     def to_dict(self) -> dict:
#         return {
#             "to": self.to,
#             "value": self.value,
#             "data": self.data
#         }

RESERVED = Schema(
    {
        Optional("features"): int, # int.
        Optional("unused"): [int] # list of int.
    },
    required=True,
    extra=REMOVE_EXTRA
)

# class Reserved():
#     ''' Reserved type.
#     Mark the transaction body if the new supplement features are used.
#     '''

#     def __init__(
#             self,
#             features: int = None,
#             unused: List[int] = None):

#         self.features = features
#         self.unused = unused

#     def to_dict(self) -> dict:
#         return {
#             "features": self.features,
#             "unused": self.unused
#         }

BODY = Schema(
    {
        "chainTag": int,
        "blockRef": str,
        "expiration": int,
        "clauses": [CLAUSE],
        "gasPriceCoef": int,
        "gas": Any(str, int),
        "dependsOn": Any(str, None),
        "nonce": Any(str, int),
        Optional("reserved"): RESERVED
    },
    required=True,
    extra=REMOVE_EXTRA
)

# class Body():
#     ''' Body type.
#     Consists of the structure of the body of a transaction.
#     '''

#     def __init__(
#             self,
#             chain_tag: int,
#             block_ref: str,
#             expiration: int,
#             clauses: List[Clause],
#             gas_price_coef: int,
#             gas: Union[str, int],
#             depends_on: Union[str, None],
#             nonce: Union[str, int],
#             reserved: Optional[Reserved] = None):

#         self.chain_tag = chain_tag
#         self.block_ref = block_ref
#         self.expiration = expiration
#         self.clauses = clauses
#         self.gas_price_coef = gas_price_coef
#         self.gas = gas
#         self.depends_on = depends_on
#         self.nonce = nonce
#         self.reserved = reserved

#     def to_dict(self) -> dict:
#         d = {
#             "chainTag": self.chain_tag,
#             "blockRef": self.block_ref,
#             "expiration": self.expiration,
#             "clauses": [x.to_dict() for x in self.clauses],
#             "gasPriceCoef": self.gas_price_coef,
#             "gas": self.gas,
#             "dependsOn": self.depends_on,
#             "nonce": self.nonce
#         }

#         if self.reserved:
#             r = {}
#             if self.reserved.features:
#                 r["features"] = self.reserved.features

#             if self.reserved.unused:
#                 r["unused"] = self.reserved.unused

#             d['reserved'] = r

#         return d


def data_gas(data: str) -> int:
    '''
    Calculate the gas the data will consume.

    Parameters
    ----------
    data : str
        '0x...' style hex string.
    '''
    Z_GAS = 4
    NZ_GAS = 68

    sum_up = 0
    for x in range(2, len(data), 2):
        if data[x] == '0' and data[x+1] == '0':
            sum_up += Z_GAS
        else:
            sum_up += NZ_GAS

    # print('sum_up', sum_up)
    return sum_up


def intrinsic_gas(clauses: List) -> int:
    '''
    Calculate roughly the gas from a list of clauses.

    Parameters
    ----------
    clauses : List
        A list of clauses (in dict format).

    Returns
    -------
    int
        The sum of gas.
    '''
    TX_GAS = 5000
    CLAUSE_GAS = 16000
    CLAUSE_CONTRACT_CREATION = 48000

    if len(clauses) == 0:
        return TX_GAS + CLAUSE_GAS

    sum_total = 0
    sum_total += TX_GAS

    for clause in clauses:
        clause_sum = 0
        if clause['to']:  # contract create.
            clause_sum += CLAUSE_GAS
        else:
            clause_sum += CLAUSE_CONTRACT_CREATION
        clause_sum += data_gas(clause['data'])

        sum_total += clause_sum

    return sum_total


class Transaction():
    # The reserved feature of delegated (vip-191) is 1.
    DELEGATED_MASK = 1

    def __init__(self, body: dict):
        ''' Construct a transaction from a given body. '''
        self.body = BODY(body)
        self.signature = None

    def _encode_reserved(self) -> List:
        r = self.body.get('reserved', None)
        if not r:
            reserved = {"features": None, "unused": None}
        else:
            reserved = self.body['reserved']

        f = reserved.get('features') or 0
        l = reserved.get('unused') or []
        m_list = [FeaturesKind.serialize(f)] + l

        # While some elements in the m_list is b'' or '',
        # Then just right strip those '' from the list.
        length_list = [len(x) for x in m_list]

        right_most_none_empty = None
        for i in range(len(length_list) - 1, -1, -1):
            if length_list[i] != 0:
                right_most_none_empty = i
                break

        if right_most_none_empty is None:  # not found the right most none-empty string item
            return []

        return_list = []
        for y in range(0, right_most_none_empty + 1):
            return_list.append(m_list[y])

        return return_list

    def get_signing_hash(self, delegate_for: str = None) -> bytes:
        reserved_list = self._encode_reserved()
        _temp = deepcopy(self.body)
        _temp.update({"reserved": reserved_list})
        buff = ComplexCodec(UnsignedTxWrapper).encode(_temp)
        h, _ = blake2b256([buff])

        if delegate_for:
            if not address.is_address(delegate_for):
                raise Exception("delegate_for should be an address type.")
            x, _ = blake2b256([h, bytes.fromhex(delegate_for[2:])])
            return x

        return h

    def get_intrinsic_gas(self) -> int:
        ''' Get the rough gas this tx will consume'''
        return intrinsic_gas(self.body['clauses'])

    def get_signature(self) -> Union[None, bytes]:
        ''' Get the signature of current transaction.'''
        return self.signature

    def set_signature(self, sig: bytes):
        ''' Set the signature '''
        self.signature = sig

    def get_origin(self) -> Union[None, str]:
        if not self._signature_valid():
            return None

        try:
            my_sign_hash = self.get_signing_hash()
            pub_key = secp256k1.recover(
                my_sign_hash, self.get_signature()[0:65])
            return '0x' + address.public_key_to_address(pub_key).hex()
        except:
            return None

    def get_delegator(self) -> Union[None, str]:
        if not self.is_delegated():
            return None

        if not self._signature_valid():
            return None

        origin = self.get_origin()
        if not origin:
            return None

        try:
            my_sign_hash = self.get_signing_hash(origin)
            pub_key = secp256k1.recover(
                my_sign_hash, self.get_signature()[65:])
            return '0x' + address.public_key_to_address(pub_key).hex()
        except:
            return None

    def is_delegated(self):
        ''' Check if this transaction is delegated.'''
        if not self.body.get('reserved'):
            return False

        if not self.body.get('reserved').get('features'):
            return False

        return self.body['reserved']['features'] & self.DELEGATED_MASK == self.DELEGATED_MASK

    def _signature_valid(self) -> bool:
        if self.is_delegated():
            expected_sig_len = 65 * 2
        else:
            expected_sig_len = 65

        if not self.get_signature():
            return False
        else:
            return len(self.get_signature()) == expected_sig_len

    def get_id(self) -> Union[None, str]:
        if not self._signature_valid():
            return None
        try:
            my_sign_hash = self.get_signing_hash()
            pub_key = secp256k1.recover(
                my_sign_hash, self.get_signature()[0:65])
            origin = address.public_key_to_address(pub_key)
            return '0x' + blake2b256([my_sign_hash, origin])[0].hex()
        except:
            return None

    def encode(self):
        ''' Encode the tx into bytes '''
        reserved_list = self._encode_reserved()
        if self.signature:
            temp = deepcopy(self.body)
            temp.update({
                'reserved': reserved_list,
                'signature': self.signature
            })
            return ComplexCodec(SignedTxWrapper).encode(temp)
        else:
            temp = deepcopy(self.body)
            temp.update({
                'reserved': reserved_list
            })
            return ComplexCodec(UnsignedTxWrapper).encode(temp)

    @staticmethod
    def decode(raw: bytes, unsigned: bool):
        ''' Return a Transaction type instance '''
        body = None
        sig = None

        if unsigned:
            body = ComplexCodec(UnsignedTxWrapper).decode(raw)
        else:
            decoded = ComplexCodec(SignedTxWrapper).decode(raw)
            sig = decoded['signature']  # bytes
            del decoded['signature']
            body = decoded

        r = body.get('reserved', [])  # list of bytes
        if len(r) > 0:
            if len(r[-1]) == 0:
                raise Exception('invalid reserved fields: not trimmed.')

            features = FeaturesKind.deserialize(r[0])
            body['reserved'] = {
                'features': features
            }
            if len(r) > 1:
                body['reserved']['unused'] = r[1:]
        else:
            del body['reserved']

        # Now body is a "dict", we try to check if it is in good shape.

        # Check if clause is in good shape.
        _clauses = []
        for each in body['clauses']:
            _clauses.append( CLAUSE(each) )
        body['clauses'] = _clauses
        
        # Check if reserved is in good shape.
        _reserved = None
        if body.get('reserved'):
            _reserved = RESERVED(body['reserved'])
            body['reserved'] = _reserved

        tx = Transaction(body)

        if sig:
            tx.set_signature(sig)

        return tx

    def __eq__(self, other):
        ''' Compare two tx to be the same? '''
        flag_1 = (self.signature == other.signature)
        flag_2 = (self.body == other.body)
        return flag_1 and flag_2